@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2021, University of Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -57,7 +57,7 @@
 int		considerGarbageCollect(Stack s);
 void		call_tune_gc_hook(void);
 int		garbageCollect(gc_reason_t reason);
-word		pl_garbage_collect(term_t d);
+foreign_t	pl_garbage_collect(term_t d);
 gc_stat *	last_gc_stats(gc_stats *stats);
 Word		findGRef(int n);
 size_t		nextStackSizeAbove(size_t n);
@@ -79,12 +79,6 @@ QueryFrame	queryOfFrame(LocalFrame fr);
 void		mark_active_environment(struct bit_vector *active,
 					LocalFrame fr, Code PC);
 void		unmark_stacks(LocalFrame fr, Choice ch, uintptr_t mask);
-#if defined(O_DEBUG) || defined(SECURE_GC) || defined(O_MAINTENANCE)
-word		checkStacks(void *vm_state);
-bool		scan_global(int marked);
-char *		print_addr(Word p, char *buf);
-char *		print_val(word w, char *buf);
-#endif
 void		blockGC(int flags);	/* disallow garbage collect */
 void		unblockGC(int flags);	/* re-allow garbage collect */
 
@@ -102,29 +96,35 @@ void		unblockGC(int flags);	/* re-allow garbage collect */
 		 *******************************/
 
 #define ensureLocalSpace_ex(bytes) LDFUNC(ensureLocalSpace_ex, bytes)
-static inline int
+static inline bool
 ensureLocalSpace_ex(DECL_LD size_t bytes)
 { int rc;
 
-  if ( likely(addPointer(lTop, bytes) <= (void*)lMax) )
-    return TRUE;
+  if ( likely(hasLocalSpace(bytes)) )
+    return true;
 
-  if ( (rc=growLocalSpace(bytes, ALLOW_SHIFT)) == TRUE )
-    return TRUE;
+  if ( (rc=growLocalSpace(bytes, ALLOW_SHIFT)) == true )
+    return true;
 
   return raiseStackOverflow(rc);
 }
 
-#define ensureStackSpace_ex(gcells, tcells, flags) LDFUNC(ensureStackSpace_ex, gcells, tcells, flags)
-static inline int
+#define ensureStackSpace_ex(gcells, tcells, flags) \
+	LDFUNC(ensureStackSpace_ex, gcells, tcells, flags)
+
+static inline bool
 ensureStackSpace_ex(DECL_LD size_t gcells, size_t tcells, int flags)
-{ gcells += BIND_GLOBAL_SPACE;
-  tcells += BIND_TRAIL_SPACE;
+{ int rc;
 
-  if ( likely(gTop+gcells <= gMax) && likely(tTop+tcells <= tMax) )
-    return TRUE;
+  if ( hasStackSpace(gcells, tcells) )
+    return true;
 
-  return f_ensureStackSpace(gcells, tcells, flags);
+  if ( (rc=f_ensureStackSpace(gcells+BIND_GLOBAL_SPACE,
+			      tcells+BIND_TRAIL_SPACE,
+			      flags)) == true )
+    return true;
+
+  return raiseStackOverflow(rc);
 }
 
-#endif /*_PL_GC_H*/
\ No newline at end of file
+#endif /*_PL_GC_H*/
